import os
import sys
import datetime
import pytest
import pandas as pd


test_year = 2023

def setup():
  code_dir = os.path.join(os.path.abspath('..'), 'src')
  sys.path.append(code_dir)
  from _pytest.assertion import truncate
  truncate.DEFAULT_MAX_LINES = 9999
  truncate.DEFAULT_MAX_CHARS = 9999  

setup()

from markets_insights.core.column_definition import BaseColumns, BasePriceColumns
from markets_insights.datareader.data_reader import DataReader
from markets_insights.core.core import DateFilter, DatePartsFilter, IdentifierFilter, TypeHelper, MarketDaysHelper

class PresetDates:
  year_start: datetime.datetime = MarketDaysHelper.get_this_or_next_market_day(datetime.date(year=test_year, month=1, day=1))
  year_end: datetime.datetime = MarketDaysHelper.get_this_or_previous_market_day(datetime.date(year=test_year, month=12, day=31))
  dec_start: datetime.datetime = MarketDaysHelper.get_this_or_next_market_day(datetime.date(year=test_year, month=12, day=1))
  dec_end: datetime.datetime = year_end
  q4_start: datetime.datetime = MarketDaysHelper.get_this_or_next_market_day(datetime.date(year=test_year, month=10, day=1))
  q4_end: datetime.datetime = year_end
  
class PresetFilters:
  nifty50 = IdentifierFilter("NIFTY 50")
  date_year_end = DateFilter(PresetDates.year_end)
  year = DatePartsFilter(year=2023)

class Presets:
  cols_to_test = [
    BaseColumns.Identifier,
    BaseColumns.Date,
    BaseColumns.Close,
    BaseColumns.Open,
    BaseColumns.High,
    BaseColumns.Turnover,
    BaseColumns.Volume
  ]
  filters = PresetFilters
  dates = PresetDates

def check_cols_present(data: pd.DataFrame, cols_to_test: [int], case_identifier: str):
  for col_name in cols_to_test:
    if not col_name in data.columns:
      assert pytest.fail(f"column {col_name} not present in {case_identifier} data")

def check_base_cols_present(data: pd.DataFrame, case_identifier: str):
  check_cols_present(data, Presets.cols_to_test, case_identifier)

def check_price_cols_present(data: pd.DataFrame, period: str):
  check_cols_present(
                      data, 
                      [col_template.substitute({"period": period}) for col_template in TypeHelper.get_class_static_values(BasePriceColumns)],
                      period
                    )

def check_col_values(data: pd.DataFrame, col_value_pairs: dict):
  for col_name in col_value_pairs:
    if not float(data[col_name].values[0]) == pytest.approx(col_value_pairs[col_name]):
      assert pytest.fail(f"column `{col_name}` value check: {data[col_name].values[0]} <> {col_value_pairs[col_name]}")


import pytest
import pandas as pd

from helper import (
  check_col_values,
  setup, 
  check_base_cols_present,
  Presets
)

setup()

import markets_insights
from markets_insights.core.column_definition import AggregationPeriods, BaseColumns, CalculatedColumns
from markets_insights.datareader.data_reader import NseIndicesReader
from markets_insights.dataprocess.data_processor import (
  HistoricalDataProcessOptions,
  HistoricalDataProcessor,
  HistoricalDataset
)

def test_historical_data_processor_without_options():
  processor = HistoricalDataProcessor()
  result: HistoricalDataset = processor.process(NseIndicesReader(), {
    'from_date': Presets.dates.dec_start,
    'to_date': Presets.dates.dec_end
  })
  check_base_cols_present(result.get_daily_data(), "Daily")
  assert result.get_daily_data().shape[0] == 2146

  check_base_cols_present(result.get_monthly_data(), AggregationPeriods.Monthly)
  assert result.get_monthly_data().shape[0] == 108

  check_base_cols_present(result.get_annual_data(), AggregationPeriods.Annual)
  assert result.get_monthly_data().shape[0] == 108
  

@pytest.mark.parametrize("options", [
  HistoricalDataProcessOptions(include_monthly_data=False),
  HistoricalDataProcessOptions(include_annual_data=False),
  HistoricalDataProcessOptions(include_monthly_data=False, include_annual_data=False)
])
def test_historical_data_processor_with_options(options: HistoricalDataProcessOptions):
  processor = HistoricalDataProcessor(options)
  result = processor.process(NseIndicesReader(), {
    'from_date': Presets.dates.dec_start,
    'to_date': Presets.dates.dec_end
  })
  check_base_cols_present(result.get_daily_data(), "Daily")

  if options.include_monthly_data:
    check_base_cols_present(result.get_monthly_data(), AggregationPeriods.Monthly)
  else:
    assert result.get_monthly_data() is None

  if options.include_annual_data:
    check_base_cols_present(result.get_annual_data(), AggregationPeriods.Annual)
  else:
    assert result.get_annual_data() is None

def test_historical_data_processor_monthly_aggregration():
  processor = HistoricalDataProcessor()
  result = processor.process(NseIndicesReader(), {
    'from_date': Presets.dates.dec_start,
    'to_date': Presets.dates.dec_end
  })

  check_col_values(
    data = result.get_monthly_data().query(str(Presets.filters.nifty50 & Presets.filters.date_year_end)),
    col_value_pairs = {
      BaseColumns.Open: 20194.1,
      BaseColumns.High: 21801.45,
      BaseColumns.Low: 20183.7,
      BaseColumns.Close: 21731.4,
      BaseColumns.Turnover: 5752296300000.0
    }
  )

def test_historical_data_processor_annual_aggregration():
  processor = HistoricalDataProcessor()
  result = processor.process(NseIndicesReader(), {
    'from_date': Presets.dates.year_start,
    'to_date': Presets.dates.year_end
  })

  check_col_values(
    data = result.get_annual_data().query(str(Presets.filters.nifty50 & Presets.filters.date_year_end)),
    col_value_pairs = {
      BaseColumns.Open: 18131.7,
      BaseColumns.High: 21801.45,
      BaseColumns.Low: 16828.35,
      BaseColumns.Close: 21731.4,
      BaseColumns.Turnover: 58075646100000.0
    }
  )


from datetime import date
import os
import sys


code_dir = os.path.join(os.path.abspath('.'), 'src')
os.chdir(code_dir)
sys.path.append(code_dir)

from markets_insights.dataprocess.data_processor import HistoricalDataProcessor
from markets_insights.datareader.data_reader import NseIndicesReader

processor = HistoricalDataProcessor()
result = processor.process(NseIndicesReader(), {
  'from_date': date(2023, 1, 1),
  'to_date': date(2023, 12, 31)
})


from string import Template

class AggregationPeriods:
  Monthly = "Monthly"
  Annual = "Annual"

class BasePriceColumns:
  Open = 'Open'
  High = 'High'
  Low = 'Low'
  Close = 'Close'

class BaseColumns(BasePriceColumns):
  Identifier = 'Identifier'
  PreviousClose = 'PreviousClose'
  Date = 'Date'
  Turnover: str = 'Turnover'
  Volume: str = 'Volume'

PeriodAggregateColumnTemplate: Template = Template("$period $col_name")

class CalculatedColumnsBase:
  Month = 'Month'
  MonthNo = 'MonthNo'
  Year = 'Year'
  Day = 'Day'
  Vwap = 'Vwap'
  PriceCrossedAboveVwap = 'PriceCrossedAboveVwap'

class CalculatedColumns (CalculatedColumnsBase):
  Change = 'Change'
  IsInDerivatives = 'IsInDerivatives'
  ClosePriceDiff = 'ClosePriceDiff'
  Gain = 'Gain'
  Loss = 'Loss'
  AvgGain = 'AvgGain'
  AvgLoss = 'AvgLoss'
  RelativeStrength = 'Rs'
  RelativeStrengthIndex = 'Rsi'
  RsiCrossedAbove = 'RsiCrossedAbove'
  RsiCrossedBelow = 'RsiCrossedBelow'
  StochRsi_K = 'StochRsi_K'
  StochRsi_D = 'StochRsi_D'
  StochRsi_KCrossedAbove = 'StochRsi_KCrossedAbove'
  StochRsi_KCrossedBelow = 'StochRsi_KCrossedBelow'
  StochRsi_DCrossedAbove = 'StochRsi_DCrossedAbove'
  StochRsi_DCrossedBelow = 'StochRsi_DCrossedBelow'

class DerivativesCalculatedColumns (CalculatedColumnsBase):
  CloseToPrevCloseChangePerc: str = 'CloseToPrevCloseChangePerc'
  OpenToPrevCloseChangePerc: str = 'OpenToPrevCloseChangePerc'
  PreviousCloseAmount: str = 'PreviousCloseAmount'
  OpenAmount: str = 'OpenAmount'
  CloseAmount: str = 'CloseAmount'
  HighAmount: str = 'HighAmount'
  LowAmount: str = 'LowAmount'
  AmountDiffOpenToPrevClose: str = 'AmountDiffOpenToPrevClose'
  AmountDiffCloseToPrevClose: str = 'AmountDiffCloseToPrevClose'
  AmountDiffOpenToPrevClose: str = 'AmountDiffOpenToPrevClose'
  ClosestStrike: str = 'ClosestStrike'
  LotSize: str = 'LotSize'

class DerivativesBaseColumns(BaseColumns):
  AveragePrice: str = 'AvrgPric'
  OptionType: str = 'OptionType'
  ExpiryDate: str = 'ExpiryDate'
  StrikePrice: str = 'StrkPric'
  OpenInterest: str = 'OpenInterest'
  OiChangePct: float = 'OiChangePerc'



  
from markets_insights.core.core import Instrumentation
from markets_insights.core.column_definition import BaseColumns, CalculatedColumns
import pandas as pd
import pandas_ta as ta

class CalculationWorker:
  def add_calculated_columns(self, df):
    raise NotImplementedError("add_calculated_fields")

class CalculationPipeline:
  _pipeline: list
  
  def __init__(self, workers = None):
    if workers is not None:
      self._pipeline = workers
    else:
      self._pipeline = []

  def add_calculation_worker(self, worker: CalculationWorker):
    self._pipeline.append(worker)

  def run(self, data):
    for worker in self._pipeline:
      result = worker.add_calculated_columns(data)
      if result is not None:
        data = result
    return data
  
class ValueCrossedAboveFlagWorker (CalculationWorker):
  def __init__(self, value_column, value):
    self._value_column = value_column
    self._value = value
    self._column_name = f'{value_column}CrossedAbove'

  @Instrumentation.trace(name="ValueCrossedAboveFlagWorker")
  def add_calculated_columns(self, data):
    identifier_grouped_data = data.groupby(BaseColumns.Identifier)
    data[self._column_name] = identifier_grouped_data[self._value_column].transform(lambda x: 
      (x.shift(1) < self._value) & (x >= self._value)
    )

class ValueCrossedBelowFlagWorker (CalculationWorker):
  def __init__(self, value_column, value):
    self._value_column = value_column
    self._value = value
    self._column_name = f'{value_column}CrossedBelow'

  @Instrumentation.trace(name="ValueCrossedBelowFlagWorker")
  def add_calculated_columns(self, data):
    identifier_grouped_data = data.groupby(BaseColumns.Identifier)
    data[self._column_name] = identifier_grouped_data[self._value_column].transform(lambda x: 
      (x.shift(-1) > self._value) & (x <= self._value)
    )

class PriceCrossedAboveValueFlagWorker (CalculationWorker):
  def __init__(self, value_column):
    self._value_column = value_column
    self._column_name = f'PriceCrossedAbove{value_column}'

  @Instrumentation.trace(name="PriceCrossedAboveValueFlagWorker")
  def add_calculated_columns(self, data):
    data[self._column_name] = (data[BaseColumns.Close] >= data[self._value_column]) & (data[BaseColumns.PreviousClose] < data[self._value_column])

class PriceCrossedBelowValueFlagWorker (CalculationWorker):
  def __init__(self, value_column):
    self._value_column = value_column
    self._column_name = f'PriceCrossedBelow{value_column}'

  @Instrumentation.trace(name="PriceCrossedBelowValueFlagWorker")
  def add_calculated_columns(self, data):
    data[self._column_name] = (data[BaseColumns.Close] < data[self._value_column]) & (data[BaseColumns.PreviousClose] >= data[self._value_column])

class DatePartsCalculationWorker (CalculationWorker):
  @Instrumentation.trace(name="DatePartsCalculationWorker")
  def add_calculated_columns(self, data):
    data[CalculatedColumns.Year] = data[BaseColumns.Date].dt.year
    data[CalculatedColumns.MonthNo] = data[BaseColumns.Date].dt.month
    data[CalculatedColumns.Month] = data[BaseColumns.Date].dt.strftime('%b')
    data[CalculatedColumns.Day] = data[BaseColumns.Date].dt.strftime('%A')

class SmaCalculationWorker (CalculationWorker):
  def __init__(self, time_window):
    self._time_window = time_window
    self._column_name = 'Sma' + str(self._time_window)

  @Instrumentation.trace(name="SMACalculationWorker")
  def add_calculated_columns(self, data):
    identifier_grouped_data = data.groupby(BaseColumns.Identifier)
    data[self._column_name] = identifier_grouped_data[BaseColumns.Close].transform(
        lambda x: x.rolling(self._time_window).mean()
      )
    
class StdDevCalculationWorker (CalculationWorker):
  def __init__(self, time_window: int = 200):
    self._time_window = time_window
    self._column_name = 'StdDev' + str(self._time_window)

  @Instrumentation.trace(name="StdDevCalculationWorker")
  def add_calculated_columns(self, data):
    identifier_grouped_data = data.groupby(BaseColumns.Identifier)
    data[self._column_name] = identifier_grouped_data[BaseColumns.Close].transform(
        lambda x: x.rolling(self._time_window).std()
      )
    
class BollingerBandCalculationWorker (CalculationWorker):
  def __init__(self, time_window: int = 200, deviation: int = 2):
    self._time_window = time_window
    self._deviation = deviation
    self._column_name_upper_band = 'Bb' + str(self._time_window) + 'Dev' + str(deviation) + 'Upper'
    self._column_name_lower_band = 'Bb' + str(self._time_window) + 'Dev' + str(deviation) + 'Lower'

  @Instrumentation.trace(name="StdDevCalculationWorker")
  def add_calculated_columns(self, data):
    data[self._column_name_upper_band] = data['Sma' + str(self._time_window)] + (data['StdDev' + str(self._time_window)] * self._deviation)
    data[self._column_name_lower_band] = data['Sma' + str(self._time_window)] - (data['StdDev' + str(self._time_window)] * self._deviation)
    
class RsiOldCalculationWorker(CalculationWorker):
  def __init__(self, time_window):
    self._time_window = time_window
    self._column_name = 'Rsi' + str(self._time_window)

  def calculate_wsm_average(self, raw_data, data, avg_col_name, abs_col_name):
    step = 1
    for i, row in enumerate(data[avg_col_name].iloc[self._time_window + step:]):
      raw_data.at[data.index[i + self._time_window + step], avg_col_name] =\
        (data[avg_col_name].iloc[i + self._time_window] *
        (self._time_window - step) +
        data[abs_col_name].iloc[i + self._time_window + step])\
        / self._time_window

  @Instrumentation.trace(name="RsiOldCalculationWorker")
  def add_calculated_columns(self, data):
    identifier_grouped_data = data.groupby(BaseColumns.Identifier)
    data[CalculatedColumns.ClosePriceDiff] = identifier_grouped_data[BaseColumns.Close].transform(lambda x: x.diff(1))
    data[CalculatedColumns.Gain] = identifier_grouped_data[CalculatedColumns.ClosePriceDiff].transform(lambda x: x.clip(lower=0).round(2))
    data[CalculatedColumns.Loss] = identifier_grouped_data[CalculatedColumns.ClosePriceDiff].transform(lambda x: x.clip(upper=0).abs().round(2))

    data[CalculatedColumns.ClosePriceDiff].fillna(0)
    data[CalculatedColumns.Gain].fillna(0)
    data[CalculatedColumns.Loss].fillna(0)

    # Get initial Averages
    data[CalculatedColumns.AvgGain] = identifier_grouped_data[CalculatedColumns.Gain].transform(lambda x: 
                                        x.rolling(window=self._time_window, min_periods=self._time_window).mean())
    data[CalculatedColumns.AvgLoss] = identifier_grouped_data[CalculatedColumns.Loss].transform(lambda x: 
                                        x.rolling(window=self._time_window, min_periods=self._time_window).mean())
    
    for identifier in data[BaseColumns.Identifier].unique():
      self.calculate_wsm_average(data, identifier_grouped_data.get_group(identifier), CalculatedColumns.AvgGain, CalculatedColumns.Gain) 
      self.calculate_wsm_average(data, identifier_grouped_data.get_group(identifier), CalculatedColumns.AvgLoss, CalculatedColumns.Loss)
    
    data[CalculatedColumns.RelativeStrength] = data[CalculatedColumns.AvgGain] / data[CalculatedColumns.AvgLoss]
    data[CalculatedColumns.RelativeStrengthIndex] = 100 - (100 / (1.0 + data[CalculatedColumns.RelativeStrength]))

class RsiCalculationWorker(CalculationWorker):
  def __init__(self, time_window):
    self._time_window = time_window
    self._column_name = 'Rsi' + str(self._time_window)

  def calculate_rsi(self, group):
    group[CalculatedColumns.RelativeStrengthIndex] = ta.rsi(group[BaseColumns.Close])
    return group

  @Instrumentation.trace(name="RsiCalculationWorker")
  def add_calculated_columns(self, data):
    result = data.groupby([BaseColumns.Identifier], group_keys=True).apply(self.calculate_rsi)
    return result.reset_index(drop=True)
  
class StochRsiCalculationWorker(CalculationWorker):
  def __init__(self, time_window: int = 14):
    self._time_window = time_window

  def calculate_stoch_rsi(self, group: pd.DataFrame):
    data = ta.stochrsi(group[BaseColumns.Close], window=14, smooth1=3, smooth2=3)
    if data is not None:
      group[CalculatedColumns.StochRsi_K] = data['STOCHRSIk_14_14_3_3']
      group[CalculatedColumns.StochRsi_D] = data['STOCHRSId_14_14_3_3']
    return group

  @Instrumentation.trace(name="StochRsiCalculationWorker")
  def add_calculated_columns(self, data):
    result = data.groupby([BaseColumns.Identifier], group_keys=True).apply(self.calculate_stoch_rsi)
    return result.reset_index(drop=True)

class VwapCalculationWorker (CalculationWorker):
  def __init__(self, time_window):
    self._time_window = time_window

  @Instrumentation.trace(name="VwapCalculationWorker")
  def add_calculated_columns(self, data):
    data[BaseColumns.Turnover] = data[BaseColumns.Turnover].replace('-', 0)
    data[BaseColumns.Volume] = data[BaseColumns.Volume].replace('-', 0)
    data[CalculatedColumns.Vwap] = data.groupby(BaseColumns.Identifier).apply(lambda x: 
      x[BaseColumns.Turnover].rolling(self._time_window).sum() / x.rolling(self._time_window)[BaseColumns.Volume].sum()
    ).reset_index(level=0, drop=True)

class LowestPriceInNextNDaysCalculationWorker (CalculationWorker):
  def __init__(self, time_window):
    self._time_window = time_window
    self._val_column_name = f'LowestPriceInNext{str(self._time_window)}Days'
    self._perc_column_name = f'HighestPercFallInNext{str(self._time_window)}Days'

  @Instrumentation.trace(name="LowestPriceInNextNDaysCalculationWorker")
  def add_calculated_columns(self, data):
    identifier_grouped_data = data.groupby(BaseColumns.Identifier)
    data[self._val_column_name] = identifier_grouped_data[BaseColumns.Low].transform(lambda x: x.rolling(self._time_window).min().shift(-self._time_window))
    data[self._perc_column_name] = (data[BaseColumns.Close] - data[self._val_column_name]) / data[BaseColumns.Close] * 100


class HighestPriceInNextNDaysCalculationWorker (CalculationWorker):
  def __init__(self, time_window):
    self._time_window = time_window
    self._val_column_name = f'HighestPriceInNext{str(self._time_window)}Days'
    self._perc_column_name = f'HighestPercRiseInNext{str(self._time_window)}Days'

  @Instrumentation.trace(name="HighestPriceInNextNDaysCalculationWorker")
  def add_calculated_columns(self, data):
    identifier_grouped_data = data.groupby(BaseColumns.Identifier)
    data[self._val_column_name] = identifier_grouped_data[BaseColumns.High].transform(lambda x: x.rolling(self._time_window).max().shift(-self._time_window))
    data[self._perc_column_name] = (data[self._val_column_name] - data[BaseColumns.Close]) / data[BaseColumns.Close] * 100


import pytest
import pandas as pd

from helper import (
  check_col_values,
  setup, 
  check_cols_present,
  check_base_cols_present,
  PresetDates
)
from markets_insights.core.column_definition import BaseColumns, DerivativesBaseColumns
from markets_insights.core.core import TypeHelper

setup()

import markets_insights

from markets_insights.datareader.data_reader import (
  DataReader,
  BhavCopyReader, NseIndicesReader, NseDerivatiesReader,
  MultiDatesDataReader,
  DateRangeDataReader
)

@pytest.mark.parametrize("reader,rows", [
  (BhavCopyReader(), 1797),
  (NseIndicesReader(), 107),
  (NseDerivatiesReader(), 12450)
])
def test_single_day_reader(reader: DataReader, rows: int):
  data = reader.read(PresetDates.dec_start)
  check_base_cols_present(data, reader.name)
  assert data.shape[0] == rows

def test_derivatives_data_reader_columns():
  data = NseDerivatiesReader().read(PresetDates.dec_start)
  check_cols_present(data, [col_name for col_name in TypeHelper.get_class_static_values(DerivativesBaseColumns)], "derivatives")

def test_multi_dates_reader():
  reader = MultiDatesDataReader(NseIndicesReader())
  data = reader.read([PresetDates.year_start, PresetDates.dec_end, PresetDates.dec_start])
  check_base_cols_present(data, reader)
  assert data.shape[0] == 319

def test_date_range_reader():
  reader = DateRangeDataReader(NseIndicesReader())
  data = reader.read(from_date=PresetDates.dec_start, to_date=PresetDates.dec_end)
  check_base_cols_present(data, reader)
  assert data.shape[0] == 2146

@pytest.mark.parametrize("reader,turnover", [
  (BhavCopyReader(), 19493616.1),
  (NseIndicesReader(), 224615300000.0)
])
def test_turnover_rescaled(reader: DataReader, turnover: float):
  data = reader.read(PresetDates.dec_start)
  check_col_values(
    data = data,
    col_value_pairs = {
      BaseColumns.Turnover: turnover
    }
  )


import math
import markets_insights as mi

from urllib.request import urlopen
from zipfile import ZipFile
from markets_insights.core.column_definition import BaseColumns, DerivativesBaseColumns
from markets_insights.core.environment import EnvironmentSettings
import os
from string import Template
import pandas as pd
from urllib.error import HTTPError
from datetime import date

from markets_insights.core.core import MarketDaysHelper, Instrumentation

class ReaderOptions:
  filename = Template("")
  url_template = Template("")
  output_path_template = Template("")
  unzip_path_template = Template("")
  primary_data_path_template = Template("")
  unzip_file = True
  cutoff_date = None

  download_timeout = 5 #seconds
   
class DataReader:
  options: ReaderOptions
  
  def __init__(self):
    self.options = ReaderOptions()
    self.volume_scale: int = 1
    self.turnover_scale: int = 1
    self.name = ""

  def get_date_parts(self, for_date):
    return {
      'year': str(for_date.year),
      'month': for_date.strftime("%B").upper()[:3],
      'day': str(for_date.strftime("%d"))
    }

  def read(self, for_date) -> pd.DataFrame:
    date_parts = self.get_date_parts(for_date)
    filenames = self.get_filenames(for_date)
    output_file_path = self.options.output_path_template.substitute( **({**EnvironmentSettings.Paths, **filenames}) )
    primary_data_file_path = self.options.primary_data_path_template.substitute( **({**EnvironmentSettings.Paths, **filenames}) )

    if not os.path.exists(output_file_path):
      Instrumentation.debug(f"Downloading data for {for_date}")
      url = self.options.url_template.substitute( **({**date_parts, **filenames}) )

      Instrumentation.debug(url)
      
      urldata = urlopen(url, timeout=self.options.download_timeout)
      with open(output_file_path,'wb') as output:
        output.write(urldata.read())

    unzip_folder_path = self.options.unzip_path_template.substitute( **({**EnvironmentSettings.Paths, **filenames}) )
    if self.options.unzip_file == True and not os.path.exists(unzip_folder_path):
      self.unzip_content(output_file_path, unzip_folder_path, for_date)

    data = self.read_data_from_file(for_date, primary_data_file_path)

    column_name_mappings = self.get_column_name_mappings()
    if column_name_mappings is not None:
      data.rename(columns = column_name_mappings, inplace = True)

    data.drop_duplicates(inplace=True)

    for col_name in [BaseColumns.Volume, BaseColumns.Turnover]:
      if col_name in data.columns:
        data[col_name] = data[col_name].apply(lambda val: val if str(val).replace(".", "").isnumeric() == True else math.nan).astype(float)

    for rescaling in [(BaseColumns.Volume, self.volume_scale), (BaseColumns.Turnover, self.turnover_scale)]:
      if rescaling[1] != 1:
        data[rescaling[0]] = data[rescaling[0]] * rescaling[1]

    return self.filter_data(data)

  def unzip_content(self, output_file_path, unzip_folder_path, for_date):
    zf = ZipFile(output_file_path)
    zf.extractall(path=unzip_folder_path)
    zf.close()

  def read_data_from_file(self, for_date, primary_data_filepath):
    primary_data = pd.read_csv(primary_data_filepath)
    primary_data['Date'] = pd.to_datetime(for_date)
    return primary_data

  def get_filenames(self, for_date):
    return Exception("Not implemented!")

  def get_column_name_mappings(self):
    return None
  
  def filter_data(self, data):
    return data

class BhavCopyReader(DataReader):
  def __init__(self):
    super().__init__()
    self.name = "nse_equities"
    __base_url = "https://archives.nseindia.com/content/historical/EQUITIES/"
    self.options.url_template = Template(__base_url + "$year/$month/$download_filename")
    self.options.output_path_template = Template(f"$DataBaseDir/$RawDataDir/$BhavDataDir/$download_filename")
    self.options.unzip_path_template = Template("$DataBaseDir/$RawDataDir/$BhavDataDir/$download_filename_wo_ext")
    self.options.primary_data_path_template = Template("$DataBaseDir/$RawDataDir/$BhavDataDir/$download_filename_wo_ext/$primary_data_filename")

  def get_filenames(self, for_date):
    __formatted_date = for_date.strftime('%d%b%Y').upper()
    return {
      'download_filename_wo_ext': f"cm{__formatted_date}bhav",
      'download_filename': f"cm{__formatted_date}bhav.csv.zip",
      'primary_data_filename':  f"cm{__formatted_date}bhav.csv"
    }
  
  def get_column_name_mappings(self):
    return {
      'SYMBOL': BaseColumns.Identifier,
      'PREVCLOSE': BaseColumns.PreviousClose,
      'OPEN': BaseColumns.Open,
      'HIGH': BaseColumns.High,
      'LOW': BaseColumns.Low,
      'CLOSE': BaseColumns.Close,
      'TOTTRDQTY': BaseColumns.Volume,
      'TOTTRDVAL': BaseColumns.Turnover
    }

  def filter_data(self, data):
    return data[data['SERIES'] == 'EQ'].reset_index(drop=True)

class NseIndicesReader(DataReader):
  def __init__(self):
    super().__init__()
    self.name = "Indices"
    self.turnover_scale = math.pow(10, 7)
    __base_url = "https://archives.nseindia.com/content/indices/"
    self.options.unzip_file = False
    self.options.url_template = Template(__base_url + "$download_filename")
    self.options.output_path_template = Template("$DataBaseDir/$RawDataDir/$NseIndicesDataDir/$download_filename")
    self.options.primary_data_path_template = Template("$DataBaseDir/$RawDataDir/$NseIndicesDataDir/$download_filename")
    self.options.cutoff_date = date.fromisoformat("2013-01-01")

  def get_filenames(self, for_date):
    __formatted_date = for_date.strftime('%d%m%Y').upper()
    return {
      'download_filename': f"ind_close_all_{__formatted_date}.csv",
      'primary_data_filename':  f"ind_close_all_{__formatted_date}.csv"
    }
  
  def get_column_name_mappings(self):
    return {
      'Index Name': BaseColumns.Identifier,
      'Open Index Value': BaseColumns.Open,
      'High Index Value': BaseColumns.High,
      'Low Index Value': BaseColumns.Low,
      'Closing Index Value': BaseColumns.Close,
      'Volume': BaseColumns.Volume,
      'Turnover (Rs. Cr.)': BaseColumns.Turnover
    }

class NseDerivatiesReaderBase(DataReader):
  def filter_data(self, data):
    return data[data['OpenInterest'] > 0].reset_index(drop=True)

class NseDerivatiesReader(NseDerivatiesReaderBase):
  def __init__(self):
    super().__init__()
    self.name = "Derivatives"
    self.turnover_scale = math.pow(10, 7)
    self.options.unzip_file = False
    __base_url = "https://archives.nseindia.com/content/fo/"
    self.options.url_template = Template(__base_url + "$download_filename")
    self.options.output_path_template = Template("$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename")
    self.options.primary_data_path_template = Template("$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename")
    
  def get_filenames(self, for_date):
    #__formatted_date = for_date.strftime('%d%b%Y').upper()
    __formatted_date = for_date.strftime('%d%m%Y')
    return {
      'download_filename': f"NSE_FO_bhavcopy_{__formatted_date}.csv",
      'primary_data_filename':  f"NSE_FO_bhavcopy_{__formatted_date}.csv"
    }
  
  def get_column_name_mappings(self):
    return {
      'TckrSymb': BaseColumns.Identifier,
      'XpryDt': DerivativesBaseColumns.ExpiryDate,
      'OptnTp': DerivativesBaseColumns.OptionType,
      'PrvsClsgPric': BaseColumns.PreviousClose,
      'TtlTradgVol': BaseColumns.Volume,
      'TtlTrfVal': BaseColumns.Turnover,
      'OpnPric': BaseColumns.Open,
      'HghPric': BaseColumns.High,
      'LwPric': BaseColumns.Low,
      'ClsPric': BaseColumns.Close,
      'OpnIntrst': DerivativesBaseColumns.OpenInterest,
      'PctgChngInOpnIntrst': DerivativesBaseColumns.OiChangePct,
      'StrkPric': DerivativesBaseColumns.StrikePrice
    }

class NseDerivatiesOldReader(NseDerivatiesReaderBase):
  def __init__(self):
    super().__init__()
    self.name = "Derivatives"
    self.turnover_scale = math.pow(10, 7)
    __base_url = "https://archives.nseindia.com/content/historical/DERIVATIVES/"
    self.options.url_template = Template(__base_url + "$year/$month/$download_filename")
    self.options.output_path_template = Template("$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename")
    self.options.unzip_path_template = Template("$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename_wo_ext")
    self.options.primary_data_path_template = Template("$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename_wo_ext/$primary_data_filename")
    
  def get_filenames(self, for_date):
    __formatted_date = for_date.strftime('%d%b%Y').upper()
    return {
      'download_filename_wo_ext': f"fo{__formatted_date}bhav",
      'download_filename': f"fo{__formatted_date}bhav.csv.zip",
      'primary_data_filename':  f"fo{__formatted_date}bhav.csv"
    }
  
  def get_column_name_mappings(self):
    return {
      'SYMBOL': BaseColumns.Identifier,
      'EXPIRY_DT': DerivativesBaseColumns.ExpiryDate,
      'OPTION_TYP': DerivativesBaseColumns.OptionType,
      'PrvsClsgPric': BaseColumns.PreviousClose,
      'TtlTradgVol': BaseColumns.Turnover,
      'OPEN': BaseColumns.Open,
      'HIGH': BaseColumns.High,
      'LOW': BaseColumns.Low,
      'CLOSE': BaseColumns.Close,
      'OpnIntrst': DerivativesBaseColumns.OpenInterest,
      'PctgChngInOpnIntrst': DerivativesBaseColumns.OiChangePct,
      'STRIKE_PR': DerivativesBaseColumns.StrikePrice
    }

class MultiDatesDataReader:
  reader: DataReader

  def __init__(self, reader: DataReader):
    self.reader = reader
  
  def read(self, datelist):
    
    result = None
    for for_date in datelist:
        if MarketDaysHelper.is_open_for_day(pd.Timestamp(for_date).date()):
            try:
              data = self.reader.read(for_date)
              if result is None:
                  result = data
              else:
                  result = pd.concat([result, data], ignore_index=True).reset_index(drop=True)
            except Exception as e:
              print(e, for_date.strftime('date(%Y, %m, %d),'))
            
    return result
  
class DateRangeDataReader (DataReader):
  reader: DataReader

  def __init__(self, reader: DataReader):
    super().__init__()
    self.reader = reader
  
  def read(self, from_date, to_date):
    
    datelist = MarketDaysHelper.get_days_list_for_range(from_date, to_date)
    result = None
    for for_date in datelist:
        if MarketDaysHelper.is_open_for_day(pd.Timestamp(for_date).date()):
            try:
              data = self.reader.read(for_date)
              if result is None:
                  result = data
              else:
                  result = pd.concat([result, data], ignore_index=True).reset_index(drop=True)
            except Exception as e:
              print(e, for_date.strftime('date(%Y, %m, %d),'))
            
    return result




from markets_insights.core.column_definition import BaseColumns, CalculatedColumns, CalculatedColumnsBase
from markets_insights.core.settings import MarketDaysSettings
from markets_insights.core.environment import EnvironmentSettings
from datetime import date
import time
import pandas as pd
from dateutil.relativedelta import relativedelta
import calendar

Weekends = ['Saturday', 'Sunday']

class TypeHelper:
  def get_class_props(c_type):
    return [x for x in dir(c_type) if x.startswith('__') == False and type(getattr(c_type, x)) != 'function']
  
  def get_class_static_values(c_type):
    return [getattr(c_type, x) for x in dir(c_type) if x.startswith('__') == False and type(getattr(c_type, x)) != 'function']

  def get_class_props_for_value(c_type, val):
    props = TypeHelper.get_class_props(c_type)
    return [prop for prop in props if val in getattr(c_type, prop)]

class InstrumentationType:
  Info = 1
  Trace = 2
  Debug = 4

InstrumentationLevel = EnvironmentSettings.Development['InstrumentationLevel']

class Instrumentation:
  
  def debug(message):
    if InstrumentationLevel & InstrumentationType.Debug == InstrumentationType.Debug:
      print(message)
  
  def info(message):
    if InstrumentationLevel & InstrumentationType.Info == InstrumentationType.Info:
      print(message)
  
  def startTracing(name):
    if InstrumentationLevel & InstrumentationType.Trace == InstrumentationType.Trace:
      return Trace(name)
    else:
      return NoTrace(name)
  
  def trace(name):
    def decorator(function):
        def wrapper(*args, **kwargs):
            tracer = Instrumentation.startTracing(name)
            result = function(*args, **kwargs)
            tracer.endTracing()
            return result
        return wrapper
    return decorator

class Trace:
  def __init__(self, name):
    self.startTime = time.time()
    self.name = name

  def endTracing(self):
    print(f"{self.name} took {round(time.time() - self.startTime)} seconds")

class NoTrace:
  def __init__(self, name):
    self.name = name

  def endTracing(self):
    return

class MarketDaysHelper:
  def is_open_for_day(for_date):
    day = for_date.strftime("%A")
    if hasattr(for_date, 'date'):
      for_date = for_date.date()
    return not (day in Weekends or for_date in MarketDaysSettings.MarketHolidays)
  
  def get_this_or_next_market_day(for_date):
    current_date = for_date
    i = 1
    while i < 5:
      if not MarketDaysHelper.is_open_for_day(current_date):
        current_date = current_date + pd.Timedelta(days=1)
      else:
        return current_date
      i = i + 1
    return current_date
  
  def get_next_market_day(for_date):
    current_date = for_date + pd.Timedelta(days=1)
    i = 1
    while i < 5:
      if not MarketDaysHelper.is_open_for_day(current_date):
        current_date = current_date + pd.Timedelta(days=1)
      else:
        return current_date
      i = i + 1
    return current_date
  
  def get_this_or_previous_market_day(for_date):
    current_date = for_date
    i = 1
    while i < 5:
      if not MarketDaysHelper.is_open_for_day(current_date):
        current_date = current_date - pd.Timedelta(days=1)
      else:
        return current_date
      i = i + 1
    return current_date
  
  def get_previous_market_day(for_date):
    current_date = for_date - pd.Timedelta(days=1)
    i = 1
    while i < 5:
      if not MarketDaysHelper.is_open_for_day(current_date):
        current_date = current_date - pd.Timedelta(days=1)
      else:
        return current_date
      i = i + 1
    return current_date
  
  def get_days_list_for_range(from_date, to_date):
    return pd.date_range(from_date, to_date, freq='B').tolist()
  
  def get_monthly_expiry_dates(no_of_recent_expiries):
    expiries = []
    today = date.today()
    for i in range(1, no_of_recent_expiries):
      cur_month = date(today.year, today.month, 1) - relativedelta(months=i)
      monthly_expiry = MarketDaysHelper.get_this_or_previous_market_day(MarketDaysHelper.get_last_thursday(cur_month.year, cur_month.month))
      expiries.append(monthly_expiry)
    return expiries
  
  def get_last_thursday(year, month):
    cal = calendar.Calendar(firstweekday=calendar.MONDAY)
    thursday_number = -1
    monthcal = cal.monthdatescalendar(year, month)
    last_thursday = [day for week in monthcal for day in week if \
        day.weekday() == calendar.THURSDAY and \
        day.month == month][thursday_number]
    return last_thursday

class FilterCriteria:
  _col_to_filter: str = None
  _condition: str = None
  _condition_value = None
  
  def __init__(self, col_to_filter: str = None, condition: str = None, condition_value = None, ):
    self._col_to_filter = col_to_filter
    self._condition = condition
    self._condition_value = condition_value

  def __str__(self) -> str:
      return self.get_query()

  def get_query(self) -> str:
      return f"`{self._col_to_filter}` {self._condition} {self._condition_value}"

class FilterBase:
  _filter_criterias: [FilterCriteria]

  def __init__(self):
    self._filter_criterias = []
    
  def add_criteria(self, criteria: FilterCriteria):
    self._filter_criterias.append(criteria)

  def __str__(self) -> str:
      return f"{self.get_query()}"

  def get_query(self) -> str:
      return " & ".join(
            [
                f"{k.get_query()}"
                for k in self._filter_criterias
            ]
        )

  def __and__(self, other: object):
    new_filter = FilterBase()
    for criteria in self._filter_criterias:
      new_filter.add_criteria(criteria)
    for criteria in other._filter_criterias:
      new_filter.add_criteria(criteria)
    return new_filter

class IdentifierFilter(FilterBase):
  def __init__(self, identifier: str):
    super().__init__()
    self.add_criteria(FilterCriteria(
        col_to_filter=BaseColumns.Identifier,
        condition = "==",
        condition_value = f"'{identifier}'"
      )
    )

class DateFilter(FilterBase):
  def __init__(self, for_date: date = None):
    super().__init__()
    self.add_criteria(FilterCriteria(
        col_to_filter=BaseColumns.Date,
        condition = "==",
        condition_value = f"'{for_date}'"
      )
    )
  
class DateRangeFilter(FilterBase):
  def __init__(self, from_date: date = None, to_date: date = None):
    super().__init__()
    if from_date is not None:
      self.add_criteria(FilterCriteria(
          col_to_filter=BaseColumns.Date,
          condition = ">=",
          condition_value = f"'{from_date}'"
        )
      )

    if to_date is not None:
      self.add_criteria(FilterCriteria(
          col_to_filter=BaseColumns.Date,
          condition = "<=",
          condition_value = f"'{to_date}'"
        )
      )

class DatePartsFilter(FilterBase):
  def __init__(self, month_no: int = None, year: int = None):
    super().__init__()
    if month_no is not None:
      self.add_criteria(FilterCriteria(
          col_to_filter=CalculatedColumns.MonthNo,
          condition = "==",
          condition_value = month_no
        )
      )
    if year is not None:
      self.add_criteria(FilterCriteria(
          col_to_filter=CalculatedColumns.Year,
          condition = "==",
          condition_value = year
        )
      )


from markets_insights.core.column_definition import DerivativesBaseColumns, DerivativesCalculatedColumns
from markets_insights.core.core import Instrumentation
from markets_insights.dataprocess.data_processor import CalculationWorker
import pandas as pd

class DerivativesPriceCalculationWorker (CalculationWorker):
  @Instrumentation.trace(name="DerivativesPriceCalculationWorker")
  def add_calculated_columns(self, data):
    if DerivativesBaseColumns.PreviousClose not in data.columns.to_list():
      data[DerivativesBaseColumns.PreviousClose] = data.groupby([DerivativesBaseColumns.Identifier, DerivativesBaseColumns.OptionType, DerivativesBaseColumns.ExpiryDate, DerivativesBaseColumns.StrikePrice])[DerivativesBaseColumns.Close].transform(lambda x: x.shift(1))

    data[DerivativesCalculatedColumns.CloseToPrevCloseChangePerc] = (data[DerivativesBaseColumns.Close] / data[DerivativesBaseColumns.PreviousClose] - 1) * 100
    data[DerivativesCalculatedColumns.OpenToPrevCloseChangePerc] = (data[DerivativesBaseColumns.Open] / data[DerivativesBaseColumns.PreviousClose] - 1) * 100
    
    data[DerivativesCalculatedColumns.PreviousCloseAmount] = data[DerivativesBaseColumns.PreviousClose] * data[DerivativesCalculatedColumns.LotSize]
    data[DerivativesCalculatedColumns.OpenAmount] = data[DerivativesBaseColumns.Open] * data[DerivativesCalculatedColumns.LotSize]
    data[DerivativesCalculatedColumns.CloseAmount] = data[DerivativesBaseColumns.Close] * data[DerivativesCalculatedColumns.LotSize]
    data[DerivativesCalculatedColumns.HighAmount] = data[DerivativesBaseColumns.High] * data[DerivativesCalculatedColumns.LotSize]
    data[DerivativesCalculatedColumns.LowAmount] = data[DerivativesBaseColumns.Low] * data[DerivativesCalculatedColumns.LotSize]

    data[DerivativesCalculatedColumns.AmountDiffOpenToPrevClose] = data[DerivativesCalculatedColumns.OpenAmount] - data[DerivativesCalculatedColumns.PreviousCloseAmount]
    data[DerivativesCalculatedColumns.AmountDiffCloseToPrevClose] = data[DerivativesCalculatedColumns.CloseAmount] - data[DerivativesCalculatedColumns.PreviousCloseAmount]

class DerivativesLotSizeCalculationWorker (CalculationWorker):
  @Instrumentation.trace(name="DerivativesLotSizeCalculationWorker")
  def add_calculated_columns(self, data):
    data[DerivativesCalculatedColumns.LotSize] = data.groupby([DerivativesBaseColumns.Identifier, DerivativesBaseColumns.ExpiryDate])[DerivativesBaseColumns.OpenInterest].transform(lambda x: x.min())
  
class DerivativesLotSizeCalculationWorker_Obsolete (CalculationWorker):
  @Instrumentation.trace(name="DerivativesLotSizeCalculationWorker")
  def add_calculated_columns(self, data):
    lotsize_data = pd.read_csv("../manual_data/nse_fo_lotsize.csv")
    cur_expiry_col = lotsize_data.columns[2]
    data = pd.merge(
        data,
        lotsize_data[['SYMBOL', cur_expiry_col]],
        how="left",
        left_on=DerivativesBaseColumns.Identifier,
        right_on='SYMBOL'
    )
    data.rename(columns={cur_expiry_col: DerivativesCalculatedColumns.LotSize}, inplace=True)
    return data


from markets_insights.core.column_definition import BaseColumns, CalculatedColumns, DerivativesBaseColumns, DerivativesCalculatedColumns
from markets_insights.core.core import Instrumentation
from markets_insights.calculations.base import CalculationWorker
import pandas as pd
from datetime import datetime
from dateutil.relativedelta import relativedelta
from markets_insights.datareader.data_reader import NseDerivatiesOldReader
from markets_insights.core.core import MarketDaysHelper

class IsInDerivativesFlagCalculationWorker (CalculationWorker):
  @Instrumentation.trace(name="IsInDerivativesFlagCalculationWorker")
  def add_calculated_columns(self, data):
    from_date = data[BaseColumns.Date].min()
    to_date = data[BaseColumns.Date].max()

    all_symbols = pd.DataFrame()
    current_date = datetime(from_date.year, from_date.month, 1)
    derivatives_reader = NseDerivatiesOldReader()
    
    while current_date <= to_date:
      derivatives_symbols = derivatives_reader.read(MarketDaysHelper.get_this_or_next_market_day(current_date))[[DerivativesBaseColumns.Identifier, DerivativesBaseColumns.Date]]
      derivatives_symbols[DerivativesCalculatedColumns.MonthNo] = current_date.month
      derivatives_symbols[DerivativesCalculatedColumns.Year] = current_date.year
      derivatives_symbols.drop_duplicates(inplace=True)
      all_symbols = pd.concat([all_symbols, derivatives_symbols], ignore_index=True)
      current_date = current_date + relativedelta(months=+1)

      all_symbols.drop_duplicates(inplace=True)
      all_symbols.drop(columns=[BaseColumns.Date], inplace=True)
      all_symbols[CalculatedColumns.IsInDerivatives] = True
    
    data = pd.merge(data, all_symbols, how='left', on=[BaseColumns.Identifier, CalculatedColumns.Year, CalculatedColumns.MonthNo])
    return data


from markets_insights.core.environment import EnvironmentSettings
import os
from string import Template
import pandas as pd
from markets_insights.datareader.data_reader import DataReader, DateRangeDataReader
from markets_insights.core.column_definition import BaseColumns, BasePriceColumns, CalculatedColumns, PeriodAggregateColumnTemplate, AggregationPeriods
from markets_insights.core.core import MarketDaysHelper, Instrumentation, TypeHelper
from markets_insights.core.column_definition import BaseColumns
from markets_insights.calculations.base import HighestPriceInNextNDaysCalculationWorker, LowestPriceInNextNDaysCalculationWorker
from markets_insights.calculations.base import CalculationWorker, CalculationPipeline, \
  SmaCalculationWorker, RsiCalculationWorker, BollingerBandCalculationWorker, StochRsiCalculationWorker, \
  ValueCrossedAboveFlagWorker, ValueCrossedBelowFlagWorker, PriceCrossedAboveValueFlagWorker, PriceCrossedBelowValueFlagWorker, \
  StdDevCalculationWorker

from datetime import date
import glob
from typing import Dict

class DataProcessor:
  output_dir_template = Template("")
  filename_template = Template("$ReaderName.csv")
  manual_data_dir_template = Template("$ManualDataPath")
  options: dict

  def process(self, reader: DataReader):
    return Exception("Not implemented!")
  
  def remove_unnamed_columns(self, data: pd.DataFrame) -> pd.DataFrame:
    return data.loc[:, ~data.columns.str.match("Unnamed")]

class ManualDataProcessor(DataProcessor):

  @Instrumentation.trace(name="ManualDataProcessor.process")
  def process(self, reader: DataReader):
    dir_path = os.path.join(self.manual_data_dir_template.substitute(**EnvironmentSettings.Paths), reader.name)
    
    Instrumentation.debug(dir_path)
    csv_files = glob.glob(os.path.join(dir_path, '*.csv'))

    dfs = []
    for file in csv_files:
      df = pd.read_csv(file)
      
      filename = file.split(os.path.sep)[-1].split('.')[0]
      column_value = filename.replace('_Data', '')
      
      df['Identifier'] = column_value
      
      dfs.append(df)

    final_df = pd.concat(dfs)

    final_df[BaseColumns.Date] = pd.to_datetime(final_df[BaseColumns.Date])

    output_file = os.path.join(self.manual_data_dir_template.substitute(**EnvironmentSettings.Paths), 
                                self.filename_template.substitute(**{'ReaderName': reader.name}))
    
    final_df.to_csv(output_file, index=False)

    return final_df

class CalculationPipelineBuilder:

  def create_pipeline_for_worker(worker: CalculationWorker):
    pipeline = CalculationPipeline()
    pipeline.add_calculation_worker(worker)
    return pipeline
  
  def create_pipeline_for_workers(workers: list[CalculationWorker]):
    pipeline = CalculationPipeline()
    for worker in workers:
      pipeline.add_calculation_worker(worker)
    return pipeline
  
  def create_bb_calculation_pipeline(windows = [200], deviations = [2, 3]):
    pipeline = CalculationPipeline()
    for window in windows:
      pipeline.add_calculation_worker(StdDevCalculationWorker(time_window=window))
      for deviation in deviations:
        worker = BollingerBandCalculationWorker(window, deviation)
        pipeline.add_calculation_worker(worker)
        pipeline.add_calculation_worker(PriceCrossedBelowValueFlagWorker(worker._column_name_lower_band))
        pipeline.add_calculation_worker(PriceCrossedAboveValueFlagWorker(worker._column_name_upper_band))
    return pipeline
  
  def create_sma_calculation_pipeline(windows = [50, 100, 200]):
    pipeline = CalculationPipeline()
    for window in windows:
      worker = SmaCalculationWorker(window)
      pipeline.add_calculation_worker(worker)
      pipeline.add_calculation_worker(PriceCrossedBelowValueFlagWorker(worker._column_name))
      pipeline.add_calculation_worker(PriceCrossedAboveValueFlagWorker(worker._column_name))
    return pipeline
  
  def create_rsi_calculation_pipeline(crossing_above_flag_value = 75, crossing_below_flag_value = 30, window = 14):
    pipeline = CalculationPipeline()
    pipeline.add_calculation_worker(RsiCalculationWorker(window))
    if crossing_above_flag_value is not None:
      pipeline.add_calculation_worker(ValueCrossedAboveFlagWorker(CalculatedColumns.RelativeStrengthIndex, crossing_above_flag_value))
    if crossing_below_flag_value is not None:
      pipeline.add_calculation_worker(ValueCrossedBelowFlagWorker(CalculatedColumns.RelativeStrengthIndex, crossing_below_flag_value))
    return pipeline
  
  def create_stoch_rsi_calculation_pipeline(crossing_above_flag_value = 80, crossing_below_flag_value = 20, window = 14):
    pipeline = CalculationPipeline()
    pipeline.add_calculation_worker(StochRsiCalculationWorker(window))
    if crossing_above_flag_value is not None:
      pipeline.add_calculation_worker(ValueCrossedAboveFlagWorker(CalculatedColumns.StochRsi_K, crossing_above_flag_value))
      pipeline.add_calculation_worker(ValueCrossedAboveFlagWorker(CalculatedColumns.StochRsi_D, crossing_above_flag_value))
    if crossing_below_flag_value is not None:
      pipeline.add_calculation_worker(ValueCrossedBelowFlagWorker(CalculatedColumns.StochRsi_K, crossing_below_flag_value))
      pipeline.add_calculation_worker(ValueCrossedBelowFlagWorker(CalculatedColumns.StochRsi_D, crossing_below_flag_value))
    return pipeline
  
  def create_forward_looking_price_fall_pipeline(n_days_list):
    forward_looking_lowest_price_pipeline = CalculationPipeline()
    for n in n_days_list:
      forward_looking_lowest_price_pipeline.add_calculation_worker(LowestPriceInNextNDaysCalculationWorker(n))
    return forward_looking_lowest_price_pipeline
  
  def create_forward_looking_price_rise_pipeline(n_days_list):
    forward_looking_highest_price_pipeline = CalculationPipeline()
    for n in n_days_list:
      forward_looking_highest_price_pipeline.add_calculation_worker(HighestPriceInNextNDaysCalculationWorker(n))
    return forward_looking_highest_price_pipeline

class MultiDataCalculationPipelines:
  def __init__(self):
    self._store: Dict[str, CalculationPipeline] = {}

  def set_item(self, k: str, v: CalculationPipeline) -> None:
    self._store[k] = v
    
  def get_item(self, k: str) -> CalculationPipeline:
    return self._store[k]
  
  def run(self, data):
    for key in self._store:
      result = self._store[key].run(data)
      if result is not None:
        data = result
    return data

class HistoricalDataset():
  _daily : pd.DataFrame = None
  _identifier_grouped : pd.core.groupby.DataFrameGroupBy = None
  _monthly : pd.DataFrame = None
  _annual : pd.DataFrame = None
  
  def sanitize_data(self):
    self._daily[BaseColumns.High] = pd.to_numeric(self._daily[BaseColumns.High].replace("-", None))
    self._daily[BaseColumns.Low] = pd.to_numeric(self._daily[BaseColumns.Low].replace("-", None))

    self._daily[BaseColumns.High].fillna(self._daily[BaseColumns.Close], inplace=True)
    self._daily[BaseColumns.Low].fillna(self._daily[BaseColumns.Close], inplace=True)
    return self

  def create_identifier_grouped(self):
    self._identifier_grouped = self.create_grouped_data(BaseColumns.Identifier)
    return self
  
  def set_daily_data(self, daily: pd.DataFrame):
    self._daily = daily
    return self
  
  def set_monthly_data(self, monthly: pd.DataFrame):
    self._monthly = monthly
    return self

  def set_annual_data(self, annual: pd.DataFrame):
    self._annual = annual
    return self
  
  def create_grouped_data(self, columns):
    return self._daily.groupby(columns)
  
  def get_identifier_grouped(self):
    return self._identifier_grouped
  
  def get_daily_data(self) -> pd.DataFrame:
    return self._daily
  
  def get_monthly_data(self) -> pd.DataFrame:
    return self._monthly
  
  def get_annual_data(self) -> pd.DataFrame:
    return self._annual

  def process(self):
    self.sanitize_data()
    self.create_grouped_data()
    return self

class HistoricalDataProcessOptions:
  def __init__(self, include_monthly_data: bool = True, include_annual_data: bool = True):
    self.include_monthly_data = include_monthly_data
    self.include_annual_data = include_annual_data
  
class HistoricalDataProcessor(DataProcessor):
  options: HistoricalDataProcessOptions

  def __init__(self, options: HistoricalDataProcessOptions = HistoricalDataProcessOptions()):
    self.output_dir_template = Template("$DataBaseDir/$ProcessedDataDir/$HistoricalDataDir/")
    self.monthly_group_data_filename = Template("$ReaderName-$MonthlySuffix.csv")
    self.annual_group_data_filename = Template("$ReaderName-$AnnualSuffix.csv")
    self.historic_highs_reset_days = 60
    self.dataset: HistoricalDataset
    self.calculation_pipelines: MultiDataCalculationPipelines
    self.options = options
    
  def set_calculation_pipelines(self, pipelines):
    self.calculation_pipelines = pipelines

  def run_calculation_pipelines(self):
    daily_data = self.calculation_pipelines.run(self.dataset.get_daily_data())
    if daily_data is not None:
      self.dataset.set_daily_data(daily_data)

  @Instrumentation.trace(name="HistoricalDataProcessor.run_base_calculations")
  def run_base_calculations(self, daily_data: pd.DataFrame):
    if not BaseColumns.PreviousClose in daily_data.columns:
      daily_data[BaseColumns.PreviousClose] = daily_data.groupby(BaseColumns.Identifier)[BaseColumns.Close].transform(lambda x: 
        x.shift(-1)
      )
    for col_name in [BaseColumns.Open, BaseColumns.High, BaseColumns.Low]:
      daily_data[col_name] = daily_data.apply(lambda x: x[col_name] if str(x[col_name]).replace(".", "").isnumeric() == True else x[BaseColumns.Close], axis=1)
      daily_data[col_name] = daily_data[col_name].astype(float)

  @Instrumentation.trace(name="HistoricalDataProcessor.process")
  def process(self, reader: DataReader, options: dict) -> HistoricalDataset:
    from_date = MarketDaysHelper.get_this_or_next_market_day(options['from_date'])
    to_date = MarketDaysHelper.get_this_or_previous_market_day(options['to_date'])

    historical_data = self.get_data(reader, options, from_date, to_date)
    manual_data = self.get_manual_data(reader, options, from_date, to_date)

    #self.rename_columns(reader, historical_data)
    
    if manual_data is not None:
      historical_data = pd.concat([historical_data, manual_data], ignore_index=True).reset_index(drop=True).sort_values(BaseColumns.Date)

    historical_data[BaseColumns.Identifier] = historical_data[BaseColumns.Identifier].str.upper()

    self.run_base_calculations(historical_data)
    
    self.dataset = HistoricalDataset()
    daily_data = self.remove_unnamed_columns(pd.DataFrame(historical_data).reset_index(drop=True))
    
    self.dataset.set_daily_data(daily_data)
    self.dataset.create_identifier_grouped()
    
    output_path = self.output_dir_template.substitute(**EnvironmentSettings.Paths)

    if self.options.include_monthly_data == True:
      self.dataset.set_monthly_data(self.add_monthly_growth_calc(daily_data))
      monthly_data_filename = self.monthly_group_data_filename.substitute({**EnvironmentSettings.Paths, 'ReaderName': reader.name})
      self.dataset.get_monthly_data().to_csv(output_path + monthly_data_filename)
    
    if self.options.include_annual_data == True:
      self.dataset.set_annual_data(self.add_yearly_growth_calc(daily_data))
      annual_data_filename = self.annual_group_data_filename.substitute({**EnvironmentSettings.Paths, 'ReaderName': reader.name})
      self.dataset.get_annual_data().to_csv(output_path + annual_data_filename)
    
    return self.dataset
  
  def add_periodic_growth_calc(self, processed_data: pd.DataFrame, period: str) -> pd.DataFrame:
    Instrumentation.debug(f"Started periodic calculation for {period}")

    periodic_grouped = processed_data.groupby([BaseColumns.Identifier, period])

    processed_data[PeriodAggregateColumnTemplate.substitute({"period": period, "col_name": BaseColumns.Open})] = periodic_grouped[BaseColumns.Open].transform(lambda x: x.iloc[0])
    processed_data[PeriodAggregateColumnTemplate.substitute({"period": period, "col_name": BaseColumns.Close})] = periodic_grouped[BaseColumns.Close].transform(lambda x: x.iloc[-1])
    processed_data[PeriodAggregateColumnTemplate.substitute({"period": period, "col_name": BaseColumns.Low})] = periodic_grouped[BaseColumns.Low].transform('min')
    processed_data[PeriodAggregateColumnTemplate.substitute({"period": period, "col_name": BaseColumns.High})] = periodic_grouped[BaseColumns.High].transform('max')
    
    processed_data[PeriodAggregateColumnTemplate.substitute({"period": period, "col_name": BaseColumns.Volume})] = periodic_grouped[BaseColumns.Volume].transform('sum')
    processed_data[PeriodAggregateColumnTemplate.substitute({"period": period, "col_name": BaseColumns.Turnover})] = periodic_grouped[BaseColumns.Turnover].transform('sum')
    
    periodic_data = periodic_grouped.last().reset_index()
    
    aggregated_cols = [BaseColumns.Volume, BaseColumns.Turnover] + TypeHelper.get_class_static_values(BasePriceColumns)

    periodic_data = periodic_data[[BaseColumns.Identifier, BaseColumns.Date, period] + \
      [PeriodAggregateColumnTemplate.substitute({"period": period, "col_name": col_name}) for col_name in aggregated_cols]]
    
    cols = {}
    for col_name in aggregated_cols:
      cols[PeriodAggregateColumnTemplate.substitute({"period": period, "col_name": col_name})] = col_name
    periodic_data.rename(columns=cols, inplace=True)

    return periodic_data

  @Instrumentation.trace(name="HistoricalDataProcessor.add_yearly_growth_calc")
  def add_yearly_growth_calc(self, processed_data):
    processed_data[CalculatedColumns.Year] = processed_data[BaseColumns.Date].dt.year
    return self.add_periodic_growth_calc(processed_data, CalculatedColumns.Year)
  
  @Instrumentation.trace(name="HistoricalDataProcessor.add_monthly_growth_calc")
  def add_monthly_growth_calc(self, processed_data):
    processed_data[CalculatedColumns.Month] = processed_data[BaseColumns.Date].dt.strftime("%Y-%m")
    return self.add_periodic_growth_calc(processed_data, CalculatedColumns.Month)

  def rename_columns(self, reader: DataReader, historical_data: pd.DataFrame):
    column_name_mappings = reader.get_column_name_mappings()
    if column_name_mappings is not None:
      historical_data.rename(columns = column_name_mappings, inplace = True)

  @Instrumentation.trace(name="HistoricalDataProcessor.get_manual_data")
  def get_manual_data(self, reader: DataReader, options: dict, from_date: date, to_date: date):
    manual_data_file = os.path.join(self.manual_data_dir_template.substitute(**EnvironmentSettings.Paths), 
                                self.filename_template.substitute(**{'ReaderName': reader.name}))

    if os.path.exists(manual_data_file):
      manual_data = pd.read_csv(manual_data_file)
      manual_data[BaseColumns.Date] = pd.to_datetime(manual_data[BaseColumns.Date])
      return manual_data[manual_data[BaseColumns.Date].dt.date.between(from_date, to_date)]
    else:
      return None

  @Instrumentation.trace(name="HistoricalDataProcessor.get_data")
  def get_data(self, reader: DataReader, options: dict, from_date: date, to_date: date):
    Instrumentation.debug("Started to read data")
    output_file = os.path.join(self.output_dir_template.substitute(**EnvironmentSettings.Paths), 
                              self.filename_template.substitute(**{'ReaderName': reader.name}))
    
    if reader.options is not None and reader.options.cutoff_date is not None and from_date < reader.options.cutoff_date:
      Instrumentation.debug(f"Reading from {reader.options.cutoff_date} instead of {from_date}")
      from_date = reader.options.cutoff_date
    
    if isinstance(reader, DateRangeDataReader):
      dateRangeReader = reader
    else:
      dateRangeReader = DateRangeDataReader(reader)
    
    save_to_file = False

    if os.path.exists(output_file):
      historical_data = pd.read_csv(output_file)
      earliest = pd.Timestamp(historical_data[BaseColumns.Date].min()).date()
      latest = pd.Timestamp(historical_data[BaseColumns.Date].max()).date()
      if earliest > from_date:
        Instrumentation.info(f"Reading data from {from_date} to {earliest}")
        read_data = dateRangeReader.read(from_date, earliest)
        historical_data = pd.concat([historical_data, read_data], ignore_index=True).reset_index(drop=True)
        save_to_file = True
      
      if latest < to_date:
        Instrumentation.info(f"Reading data from {latest} to {to_date}")
        read_data = dateRangeReader.read(latest, to_date)
        historical_data = pd.concat([historical_data, read_data], ignore_index=True).reset_index(drop=True)
        save_to_file = True
    else:
      Instrumentation.info(f"Reading data from {from_date} to {to_date}")
      historical_data = dateRangeReader.read(from_date, to_date)
      save_to_file = True
      
    historical_data[BaseColumns.Date] = pd.to_datetime(historical_data[BaseColumns.Date])
    
    if save_to_file == True:
      Instrumentation.debug(f"Saving data to file: {output_file}")
      historical_data = self.remove_unnamed_columns(historical_data.sort_values(BaseColumns.Date))
      historical_data.to_csv(output_file, index=False)
    
    return historical_data[historical_data[BaseColumns.Date].dt.date.between(from_date, to_date)]


import os


class EnvironmentSettings:
    Paths = {
        'DataBaseDir': "../../_data",
        'RawDataDir': "raw",
        'ProcessedDataDir': "processed",
        'HistoricalDataDir': "historical",
        'MonthlySuffix': "monthly",
        'AnnualSuffix': "annual",
        'BhavDataDir': "bhavcopy",
        'NseDerivativesDataDir': "nse_derivatives",
        'NseIndicesDataDir': "nse_indices",
        'ManualDataPath': "../manual_data"
    }
    Development = {
        "InstrumentationLevel": 1 | 2 | 4
    }

class Environment:
    def setup(cache_data_base_path: str = None):
        if cache_data_base_path is not None:
            EnvironmentSettings.Paths['DataBaseDir'] = cache_data_base_path

        Environment._setup_cache_dirs()
    
    def _setup_cache_dirs():
        env_paths = EnvironmentSettings.Paths
        
        cur_path = f"{env_paths['DataBaseDir']}"
        if not os.path.exists(cur_path):
            os.mkdir(cur_path)

        for folder in ['RawDataDir', 'ProcessedDataDir']:
            cur_path = f"{env_paths['DataBaseDir']}/{env_paths[folder]}"
            if not os.path.exists(cur_path):
                os.mkdir(cur_path)
        
        for folder in ['BhavDataDir', 'NseIndicesDataDir', 'NseDerivativesDataDir']:
            cur_path = f"{env_paths['DataBaseDir']}/{env_paths['RawDataDir']}/{env_paths[folder]}"
            if not os.path.exists(cur_path):
                os.mkdir(cur_path)

        cur_path = f"{env_paths['DataBaseDir']}/{env_paths['ProcessedDataDir']}/{env_paths['HistoricalDataDir']}"
        if not os.path.exists(cur_path):
            os.mkdir(cur_path)

        